##############################################################
## Bias reduction and bias correction for beta regressions models with
## common precision parameter (precision covariates are not supported)
##############################################################
## Author: Ioannis Kosmidis
## Email: <ioannis at stats dot ucl dot ac dot uk>
## Latest release: 13/08/2010
## Distributed under GPL 2 or greater:
## Available at http://www.gnu.org/licenses
##############################################################
## "improveFit"
## ----------------
## Arguments:
##
## obj: a betareg object
##
## epsilon: tolerance for sum(abs(score))
##
## fisher: if TRUE fisher scoring is used, if FALSE Newton-Raphson is
## used
##
## slowit: the Fisher scoring or Newton-Raphson iteration is slowed
## down by multiplying the step with the slowit value (default
## 0.5). For slowit = 1 the original iterations are retrieved.
##
## maxit: maximum allowable number of iterations for the Fisher
## scoring or Newton-Raphson procedure.
##
## trace: if TRUE the function stops after each iteration printing the
## current value of the score functions and giving the user the
## ability to check things
##
##
## Value:
##
## a new betareg object which contains all the objects of the original
## betareg object, where the estimates have been updated with the more
## accurate ones and the fitted values have been updated with the ones
## corresponding to the more accurate estimates. In addition,
## obj$stdErrs are the estimated standard errors for the maximum
## likelihood estimators based on the expected information matrix and
## obj$scores are the values of the score functions at the more
## accurate estimates (typically much closer to zero than at the
## estimates that betareg returns)
##############################################################
## "lDersBias" and "lDers"
## -----------------------
## Arguments:
##
## obj: a betareg object (which includes output from improveFit)
##
##
## Value:
##
## both lDers and lDersBias return a list containing the maximum
## likelihood score functions (scores), the observed infromation
## matrix (ObsInfo), the expected information matrix (ExpInfo) and the
## fitted values (fitted), all evaluated at the estimates in obj.  The
## list returned by lDersBias also contains the first-order bias term
## (biases) and the adjustment to the score functions (adjustment),
## both evaluated at the estimates in obj.
##############################################################
## "BiasRed"
## -----------------------
## Arguments:
##
## obj: a betareg object (which includes output from improveFit)
##
##
## Value:
##
## a new betareg object which contains all the objects of the original
## betareg object, where the estimates have been replaced with the
## bias-corrected estimates, and the fitted values have been updated
## with the ones corresponding to the bias-corrected estimates. In
## addition, obj$stdErrs are the estimated standard errors for the
## bias-corrected estimators based on the expected information matrix
##############################################################
## "BiasRed"
## -----------------------
## Arguments:
##
## obj: a betareg object (which includes output from improveFit)
## epsilon: tolerance for sum(abs(adjscore))
##
## fisher: if TRUE modified fisher scoring is used, if FALSE quasi
## Newton-Raphson is used.
##
## slowit: the Fisher scoring or Newton-Raphson iteration is slowed
## down by multiplying the step with the slowit value (default
## 0.5). For slowit = 1 the original iterations are retrieved.
##
## maxit: maximum allowable number of iterations for the modified Fisher
## scoring or quasi Newton-Raphson procedure.
##
## trace: if TRUE the function stops after each iteration printing the
## current value of the adjusted score functions and giving the user
## the ability to check things
##
## keep.history: if TRUE the candidate estimates for the model
## parameters at each iteration are stored in the columns of a matrix
## and returned in obj$history. The first column of obj$history
## contains the estimates in obj.
##
## method: The bias-reducing adjusted score to be used; either
## "Expected" for adjustments based on the expected information matrix
## or "Observed" for adjustments based on the observed information
## matrix.
##
##
## Value: a new betareg object which contains all the objects of the
## original betareg object, where the original maximum likelihood
## estimates have been replaced with the bias-reduced ones and the
## fitted values have been updated with the ones corresponding to the
## bias-reduced estimates. In addition, obj$stdErrs are the estimated
## standard errors (based on the expected information matrix) for the
## bias-reduced estimators and obj$adjustedScores are the values of
## the adjusted score functions at the bias-reduced estimates. If
## keep.history is TRUE then obj$history contains the estimates at
## each iteration of the fitting procedure. Typically, if slowit = 1,
## fisher = FALSE and method = "Observed" the second column of
## obj$history contains the bias-corrected estimates (the same happens
## if slowit = 1, fisher = TRUE and method = "Expected"). These
## estimates might differ by a very small insignificant amount from
## the result that BiasCor gives due to numerical error from the extra
## matrix inversion and matrix multiplication that is involved in the
## calculations of BiasRed.
##############################################################

improveFit <- function(obj, epsilon = 1e-12, maxit = 100,
                       slowit = 0.5, fisher = TRUE,
                       trace = FALSE) {
  test <- TRUE
  niter <- 0
  cc <- obj$coefficients
  names1 <- c(names(cc$mean), names(cc$precision))
  pp <- length(names1)
  while (test & (niter < maxit)) {
    niter <- niter + 1
    ders <- lDers(obj)
    thetas <- c(obj$coefficients$mean, obj$coefficients$precision)
    direct <- slowit * (if (fisher) chol2inv(chol(ders$ExpInfo))
    else chol2inv(chol(ders$ObsInfo))) %*% ders$scores
    thetas <-  thetas + direct
    if (trace) {
      cat("Scores:", ders$scores, "\n")
      browser()
    }
    names(thetas) <- names1
    obj$coefficients$mean <- thetas[-pp]
    obj$coefficients$precision <- thetas[pp]
    test <- sum(abs(ders$scores)) > epsilon
  }
  if (niter == maxit) warning("Iteration Limit reached")
  obj$fitted.values <- ders$fitted
  obj$stdErrs <- sqrt(diag(vcov <- solve(ders$ExpInfo)))
  obj$vcov <- vcov
  obj$scores <- ders$scores
  obj$niter <- niter
  obj
}

lDersBias <- function(obj) {
  thetas <- obj$coefficients
  phi <- thetas$precision
  betas <- thetas$mean
  X <- model.matrix(obj)
  Y <- obj$y
  etas <- c(X %*% betas + (if (is.null(obj$offset)) 0 else obj$offset))
  mus <- obj$link$mean$linkinv(etas)
  D <- obj$link$mean$mu.eta(etas)
  lambdas <- psigamma(phi*mus, 0) - psigamma(phi, 0)
  ksis <- psigamma(phi*(1 - mus), 0) - psigamma(phi, 0)
  U <- log(Y) - lambdas
  Z <- log(1 - Y) - ksis
  delta <- -psigamma(phi, 1)
  nus <- psigamma(phi * mus, 1) + delta
  taus <- psigamma(phi * (1 - mus), 1) + delta
  kappa2 <- nus + taus - 2 * delta
  W <- phi * D^2 * kappa2
  W.X <- sqrt(W) * X
  Fbetabeta <- phi * crossprod(W.X)
  Fphibeta <- phi * colSums((mus * nus - (1 - mus) * taus - (2*mus - 1) * delta) * D * X)
  Fphiphi <- sum(mus^2 * nus + (1 - mus)^2 * taus + 2 * mus * (1 - mus) * delta)
  Ddash <- mus* (1 - mus) * (1 - 2 * mus)
  ## only if the link is logit. For other links replace Ddash above
  ## appropriately with the second derivation of the mean with respect
  ## to the linear predictors
  Ebetabeta <- phi * t(X) %*% (X * (U - Z) * Ddash)
  Ephibeta <- colSums((U - Z) * D * X)
  ExpInfo <- rbind(cbind(Fbetabeta, Fphibeta),
                  c(Fphibeta, Fphiphi))
  ObsInfo <- ExpInfo - rbind(cbind(Ebetabeta,  Ephibeta),
                             c(Ephibeta, 0))
  scores <-  c(phi * colSums((U - Z) * D * X), sum(mus * U + (1 - mus) * Z))
  kappa3 <- psigamma(phi * mus, 2) - psigamma(phi * (1 - mus), 2)
  Psi1 <-  psigamma(phi * (1 - mus), 1)
  Psi2 <-  psigamma(phi * (1 - mus), 2)
  ExpInfoInv <- solve(ExpInfo)
  PQ <- function(t) {
    if (t %in% (1:length(betas))) {
      Tt <- X[,t]
      gammagamma <-
        phi^3 * t(X) %*% (X * D^3 * kappa3 * Tt) +
          phi^2 * t(X) %*% (X * D * Ddash * kappa2 * Tt)
      gammaphi <-
        colSums(phi^2 * D^2 * (mus * kappa3 + Psi2) * Tt * X) +
          colSums(phi * D^2 * kappa2 * Tt * X)
      phiphi <- phi * sum(D * (mus^2 * kappa3 + (2 * mus - 1) * Psi2) * Tt)
    }
    else {
      gammagamma <-
        phi^2 * t(X) %*% (X * D^2 * (mus * kappa3 + Psi2)) +
          phi * t(X) %*% (X * Ddash * (mus * kappa2 - Psi1))
      gammaphi <-
        phi * colSums(D * (mus^2 * kappa3 + (2 * mus - 1) * Psi2) * X) +
          colSums(D * (mus * kappa2 - Psi1) * X)
      phiphi <- sum(mus^3 * kappa3 + (3 * mus^2 - 3 * mus + 1) * Psi2 - psigamma(phi, 2))
    }
    pq <- rbind(cbind(gammagamma, gammaphi), c(gammaphi, phiphi))
    res <- sum(diag(ExpInfoInv %*% pq))/2
    res
  }
  biases <-  - ExpInfoInv %*% (ADJ <- sapply(1:(length(betas) + 1), PQ))
  list(scores = scores, ObsInfo = ObsInfo, ExpInfo = ExpInfo, fitted = mus,
       biases = biases, adjustment = ADJ)
}

lDers <- function(obj) {
  thetas <- obj$coefficients
  phi <- thetas$precision
  betas <- thetas$mean
  X <- model.matrix(obj)
  Y <- obj$y
  etas <- c(X %*% betas + (if (is.null(obj$offset)) 0 else obj$offset))
  mus <- obj$link$mean$linkinv(etas)
  D <- obj$link$mean$mu.eta(etas)
  lambdas <- psigamma(phi*mus, 0) - psigamma(phi, 0)
  ksis <- psigamma(phi*(1 - mus), 0) - psigamma(phi, 0)
  U <- log(Y) - lambdas
  Z <- log(1 - Y) - ksis
  delta <- -psigamma(phi, 1)
  nus <- psigamma(phi * mus, 1) + delta
  taus <- psigamma(phi * (1 - mus), 1) + delta
  W <- phi * D^2 * (nus + taus - 2 * delta)
  W.X <- sqrt(W) * X
  Fbetabeta <- phi * crossprod(W.X)
  Fphibeta <- phi * colSums((mus * nus - (1 - mus) * taus - (2*mus - 1) * delta) * D * X)
  Fphiphi <- sum(mus^2 * nus + (1 - mus)^2 * taus + 2 * mus * (1 - mus) * delta)
  Ddash <- mus* (1 - mus) * (1 - 2 * mus)
  ## only if the link is logit. For other links replace Ddash above
  ## appropriately with the second derivation of the mean with respect
  ## to the linear predictors
  Ebetabeta <- phi * t(X) %*% (X * (U - Z) * Ddash)
  Ephibeta <- colSums((U - Z) * D * X)
  ExpInfo <- rbind(cbind(Fbetabeta, Fphibeta),
                  c(Fphibeta, Fphiphi))
  ObsInfo <- ExpInfo - rbind(cbind(Ebetabeta,  Ephibeta),
                             c(Ephibeta, 0))
  scores <-  c(phi * colSums((U - Z) * D * X), sum(mus * U + (1 - mus) * Z))
  list(scores = scores, ObsInfo = ObsInfo, ExpInfo = ExpInfo, fitted = mus)
}

BiasCor <- function(obj) { #obj is in the original parameterization
  cc <- obj$coefficients
  names1 <- c(names(cc$mean), names(cc$precision))
  pp <- length(names1)
  ders <- lDersBias(obj)
  obj$coefficients$mean <- obj$coefficients$mean - ders$biases[-pp]
  obj$coefficients$precision <- obj$coefficients$precision -
    ders$biases[pp]
  ders <- lDersBias(obj)
  obj$fitted.values <- ders$fitted
  obj$stdErrs <- sqrt(diag(vcov <- solve(ders$ExpInfo)))
  obj$vcov <- vcov
  obj
}

BiasRed <- function(obj, epsilon = 1e-12, maxit = 100,
                    slowit = 0.5, fisher = FALSE,
                    keep.history = FALSE, trace = FALSE,
                    method = c("Expected", "Observed")) {
  method <- match.arg(method)
  adjExpr <- switch(method,
                    "Expected" = expression(ders$adjustment),
                    "Observed" = expression(c(ders$ObsInfo %*% (ExpInfoInv %*% ders$adjustment))))
  test <- TRUE
  niter <- 0
  cc <- obj$coefficients
  names1 <- c(names(cc$mean), names(cc$precision))
  pp <- length(names1)
  if (keep.history) coefHistory <- matrix(NA, length(names1), maxit)
  while (test & (niter < maxit)) {
    niter <- niter + 1
    ders <- lDersBias(obj)
    thetas <- c(obj$coefficients$mean, obj$coefficients$precision)
    ExpInfoInv <- chol2inv(chol(ders$ExpInfo))
    adjustment <- eval(adjExpr)
    adjscores <- ders$scores + adjustment
    if (keep.history) {
      coefHistory[, niter] <- thetas
    }
    if (trace) {
      cat("Adjusted scores:", adjscores, "\n")
      browser()
    }
    direct <- slowit * (if (fisher) ExpInfoInv
    else chol2inv(chol(ders$ObsInfo))) %*% adjscores
    thetas <-  thetas + direct
    names(thetas) <- names1
    obj$coefficients$mean <- thetas[-pp]
    obj$coefficients$precision <- thetas[pp]
    test <- sum(abs(adjscores)) > epsilon
  }
  if (niter == maxit) warning("Iteration Limit reached")
  obj$fitted.values <- ders$fitted
  obj$adjustedScores <- adjscores
  obj$stdErrs <- sqrt(diag(ExpInfoInv))
  obj$vcov <- ExpInfoInv
  if (keep.history) obj$history <- coefHistory
  obj$scores <- NULL
  obj
}
